from __future__ import annotations

import warnings
from datetime import datetime
from typing import Any, Literal

from cognite.client._constants import OMITTED, Omitted
from cognite.client.data_classes.data_modeling import DirectRelationReference
from cognite.client.data_classes.data_modeling.ids import ViewId
from cognite.client.data_classes.data_modeling.instances import (
    PropertyOptions,
    TypedNode,
    TypedNodeApply,
)


class _CogniteExtractorDataProperties:
    extracted_data = PropertyOptions("extractedData")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_extraction_extensions", "CogniteExtractorData", "v1")


class CogniteExtractorDataApply(_CogniteExtractorDataProperties, TypedNodeApply):
    """This represents the writing format of Cognite extractor datum.

    It is used when data is written to CDF.

    Args:
        space (str): The space where the node is located.
        external_id (str): The external id of the Cognite extractor datum.
        extracted_data (dict | None | Omitted): Unstructured information extracted from source system
        existing_version (int | None): Fail the ingestion request if the node's version is greater than or equal to this value. If no existingVersion is specified, the ingestion will always overwrite any existing data for the node (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
        type (DirectRelationReference | tuple[str, str] | None | Omitted): Direct relation pointing to the type node.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        *,
        extracted_data: dict | None | Omitted = OMITTED,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None | Omitted = OMITTED,
    ) -> None:
        TypedNodeApply.__init__(self, space, external_id, existing_version, type)
        self.extracted_data = extracted_data


class CogniteExtractorData(_CogniteExtractorDataProperties, TypedNode):
    """This represents the reading format of Cognite extractor datum.

    It is used when data is read from CDF.

    Args:
        space (str): The space where the node is located.
        external_id (str): The external id of the Cognite extractor datum.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        extracted_data (dict | None): Unstructured information extracted from source system
        type (DirectRelationReference | None): Direct relation pointing to the type node.
        deleted_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        extracted_data: dict | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.extracted_data = extracted_data

    def as_write(self) -> CogniteExtractorDataApply:
        return CogniteExtractorDataApply(
            self.space,
            self.external_id,
            extracted_data=self.extracted_data,
            existing_version=self.version,
            type=self.type,
        )


class _CogniteExtractorFileProperties:
    source_id = PropertyOptions("sourceId")
    source_context = PropertyOptions("sourceContext")
    source_created_time = PropertyOptions("sourceCreatedTime")
    source_updated_time = PropertyOptions("sourceUpdatedTime")
    source_created_user = PropertyOptions("sourceCreatedUser")
    source_updated_user = PropertyOptions("sourceUpdatedUser")
    mime_type = PropertyOptions("mimeType")
    is_uploaded = PropertyOptions("isUploaded")
    uploaded_time = PropertyOptions("uploadedTime")
    extracted_data = PropertyOptions("extractedData")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_extraction_extensions", "CogniteExtractorFile", "v1")


class CogniteExtractorFileApply(_CogniteExtractorFileProperties, TypedNodeApply):
    """This represents the writing format of Cognite extractor file.

    It is used when data is written to CDF.

    Args:
        space (str): The space where the node is located.
        external_id (str): The external id of the Cognite extractor file.
        name (str | None | Omitted): Name of the instance
        description (str | None | Omitted): Description of the instance
        tags (list[str] | None | Omitted): Text based labels for generic use, limited to 1000
        aliases (list[str] | None | Omitted): Alternative names for the node
        source_id (str | None | Omitted): Identifier from the source system
        source_context (str | None | Omitted): Context of the source id. For systems where the sourceId is globally unique, the sourceContext is expected to not be set.
        source (DirectRelationReference | tuple[str, str] | None | Omitted): Direct relation to a source system
        source_created_time (datetime | None | Omitted): When the instance was created in source system (if available)
        source_updated_time (datetime | None | Omitted): When the instance was last updated in the source system (if available)
        source_created_user (str | None | Omitted): User identifier from the source system on who created the source data. This identifier is not guaranteed to match the user identifiers in CDF
        source_updated_user (str | None | Omitted): User identifier from the source system on who last updated the source data. This identifier is not guaranteed to match the user identifiers in CDF
        assets (list[DirectRelationReference | tuple[str, str]] | None | Omitted): List of assets this file relates to
        mime_type (str | None | Omitted): MIME type of the file
        directory (str | None | Omitted): Contains the path elements from the source (for when the source system has a file system hierarchy or similar)
        is_uploaded (bool | None | Omitted): Whether the file content has been uploaded to Cognite Data Fusion
        uploaded_time (datetime | None | Omitted): Point in time when the file upload was completed and the file was made available
        category (DirectRelationReference | tuple[str, str] | None | Omitted): Direct relation to an instance of CogniteFileCategory representing the detected categorization/class for the file
        extracted_data (dict | None | Omitted): Unstructured information extracted from source system
        existing_version (int | None): Fail the ingestion request if the node's version is greater than or equal to this value. If no existingVersion is specified, the ingestion will always overwrite any existing data for the node (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
        type (DirectRelationReference | tuple[str, str] | None | Omitted): Direct relation pointing to the type node.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        *,
        name: str | None | Omitted = OMITTED,
        description: str | None | Omitted = OMITTED,
        tags: list[str] | None | Omitted = OMITTED,
        aliases: list[str] | None | Omitted = OMITTED,
        source_id: str | None | Omitted = OMITTED,
        source_context: str | None | Omitted = OMITTED,
        source: DirectRelationReference | tuple[str, str] | None | Omitted = OMITTED,
        source_created_time: datetime | None | Omitted = OMITTED,
        source_updated_time: datetime | None | Omitted = OMITTED,
        source_created_user: str | None | Omitted = OMITTED,
        source_updated_user: str | None | Omitted = OMITTED,
        assets: list[DirectRelationReference | tuple[str, str]] | None | Omitted = OMITTED,
        mime_type: str | None | Omitted = OMITTED,
        directory: str | None | Omitted = OMITTED,
        is_uploaded: bool | None | Omitted = OMITTED,
        uploaded_time: datetime | None | Omitted = OMITTED,
        category: DirectRelationReference | tuple[str, str] | None | Omitted = OMITTED,
        extracted_data: dict | None | Omitted = OMITTED,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None | Omitted = OMITTED,
    ) -> None:
        TypedNodeApply.__init__(self, space, external_id, existing_version, type)
        self.name = name
        self.description = description
        self.tags = tags
        self.aliases = aliases
        self.source_id = source_id
        self.source_context = source_context
        self.source = DirectRelationReference.load(source) if source else source
        self.source_created_time = source_created_time
        self.source_updated_time = source_updated_time
        self.source_created_user = source_created_user
        self.source_updated_user = source_updated_user
        self.assets = [DirectRelationReference.load(a) for a in assets] if assets else assets
        self.mime_type = mime_type
        self.directory = directory
        self.is_uploaded = is_uploaded
        self.uploaded_time = uploaded_time
        self.category = DirectRelationReference.load(category) if category else category
        self.extracted_data = extracted_data

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        data = super().dump()
        properties = data["sources"][0]["properties"]

        # TODO: Remove attributes entirely in v8
        system_managed_props = ("isUploaded", "uploadedTime") if camel_case else ("is_uploaded", "uploaded_time")
        for prop in system_managed_props:
            if properties.get(prop) is None:
                properties.pop(prop)
            else:
                warnings.warn(
                    f"The property '{prop}' is system-managed and cannot be modified (but the non-null value set will "
                    "be included in the request). In next major version (v8), CogniteExtractorFileApply won't have "
                    "this property at all.",
                    UserWarning,
                )
        return data


class CogniteExtractorFile(_CogniteExtractorFileProperties, TypedNode):
    """This represents the reading format of Cognite extractor file.

    It is used when data is read from CDF.

    Args:
        space (str): The space where the node is located.
        external_id (str): The external id of the Cognite extractor file.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        name (str | None): Name of the instance
        description (str | None): Description of the instance
        tags (list[str] | None): Text based labels for generic use, limited to 1000
        aliases (list[str] | None): Alternative names for the node
        source_id (str | None): Identifier from the source system
        source_context (str | None): Context of the source id. For systems where the sourceId is globally unique, the sourceContext is expected to not be set.
        source (DirectRelationReference | None): Direct relation to a source system
        source_created_time (datetime | None): When the instance was created in source system (if available)
        source_updated_time (datetime | None): When the instance was last updated in the source system (if available)
        source_created_user (str | None): User identifier from the source system on who created the source data. This identifier is not guaranteed to match the user identifiers in CDF
        source_updated_user (str | None): User identifier from the source system on who last updated the source data. This identifier is not guaranteed to match the user identifiers in CDF
        assets (list[DirectRelationReference] | None): List of assets this file relates to
        mime_type (str | None): MIME type of the file
        directory (str | None): Contains the path elements from the source (for when the source system has a file system hierarchy or similar)
        is_uploaded (bool | None): Whether the file content has been uploaded to Cognite Data Fusion
        uploaded_time (datetime | None): Point in time when the file upload was completed and the file was made available
        category (DirectRelationReference | None): Direct relation to an instance of CogniteFileCategory representing the detected categorization/class for the file
        extracted_data (dict | None): Unstructured information extracted from source system
        type (DirectRelationReference | None): Direct relation pointing to the type node.
        deleted_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        source_id: str | None = None,
        source_context: str | None = None,
        source: DirectRelationReference | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        assets: list[DirectRelationReference] | None = None,
        mime_type: str | None = None,
        directory: str | None = None,
        is_uploaded: bool | None = None,
        uploaded_time: datetime | None = None,
        category: DirectRelationReference | None = None,
        extracted_data: dict | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.name = name
        self.description = description
        self.tags = tags
        self.aliases = aliases
        self.source_id = source_id
        self.source_context = source_context
        self.source = DirectRelationReference.load(source) if source else source
        self.source_created_time = source_created_time
        self.source_updated_time = source_updated_time
        self.source_created_user = source_created_user
        self.source_updated_user = source_updated_user
        self.assets = [DirectRelationReference.load(a) for a in assets] if assets else assets
        self.mime_type = mime_type
        self.directory = directory
        self.is_uploaded = is_uploaded
        self.uploaded_time = uploaded_time
        self.category = DirectRelationReference.load(category) if category else category
        self.extracted_data = extracted_data

    def as_write(self) -> CogniteExtractorFileApply:
        return CogniteExtractorFileApply(
            self.space,
            self.external_id,
            name=self.name,
            description=self.description,
            tags=self.tags,
            aliases=self.aliases,
            source_id=self.source_id,
            source_context=self.source_context,
            source=self.source,
            source_created_time=self.source_created_time,
            source_updated_time=self.source_updated_time,
            source_created_user=self.source_created_user,
            source_updated_user=self.source_updated_user,
            assets=self.assets,  # type: ignore[arg-type]
            mime_type=self.mime_type,
            directory=self.directory,
            is_uploaded=self.is_uploaded,
            uploaded_time=self.uploaded_time,
            category=self.category,
            extracted_data=self.extracted_data,
            existing_version=self.version,
            type=self.type,
        )


class _CogniteExtractorTimeSeriesProperties:
    is_step = PropertyOptions("isStep")
    time_series_type = PropertyOptions("type")
    source_id = PropertyOptions("sourceId")
    source_context = PropertyOptions("sourceContext")
    source_created_time = PropertyOptions("sourceCreatedTime")
    source_updated_time = PropertyOptions("sourceUpdatedTime")
    source_created_user = PropertyOptions("sourceCreatedUser")
    source_updated_user = PropertyOptions("sourceUpdatedUser")
    source_unit = PropertyOptions("sourceUnit")
    extracted_data = PropertyOptions("extractedData")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_extraction_extensions", "CogniteExtractorTimeSeries", "v1")


class CogniteExtractorTimeSeriesApply(_CogniteExtractorTimeSeriesProperties, TypedNodeApply):
    """This represents the writing format of Cognite extractor time series.

    It is used when data is written to CDF.

    Args:
        space (str): The space where the node is located.
        external_id (str): The external id of the Cognite extractor time series.
        is_step (bool): Defines whether the time series is a step series or not.
        time_series_type (Literal['numeric', 'string']): Defines data type of the data points.
        name (str | None | Omitted): Name of the instance
        description (str | None | Omitted): Description of the instance
        tags (list[str] | None | Omitted): Text based labels for generic use, limited to 1000
        aliases (list[str] | None | Omitted): Alternative names for the node
        source_id (str | None | Omitted): Identifier from the source system
        source_context (str | None | Omitted): Context of the source id. For systems where the sourceId is globally unique, the sourceContext is expected to not be set.
        source (DirectRelationReference | tuple[str, str] | None | Omitted): Direct relation to a source system
        source_created_time (datetime | None | Omitted): When the instance was created in source system (if available)
        source_updated_time (datetime | None | Omitted): When the instance was last updated in the source system (if available)
        source_created_user (str | None | Omitted): User identifier from the source system on who created the source data. This identifier is not guaranteed to match the user identifiers in CDF
        source_updated_user (str | None | Omitted): User identifier from the source system on who last updated the source data. This identifier is not guaranteed to match the user identifiers in CDF
        source_unit (str | None | Omitted): Unit as specified in the source system
        unit (DirectRelationReference | tuple[str, str] | None | Omitted): direct relation to the unit of the time series
        assets (list[DirectRelationReference | tuple[str, str]] | None | Omitted): The asset field.
        equipment (list[DirectRelationReference | tuple[str, str]] | None | Omitted): The equipment field.
        extracted_data (dict | None | Omitted): Unstructured information extracted from source system
        existing_version (int | None): Fail the ingestion request if the node's version is greater than or equal to this value. If no existingVersion is specified, the ingestion will always overwrite any existing data for the node (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
        type (DirectRelationReference | tuple[str, str] | None | Omitted): Direct relation pointing to the type node.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        *,
        is_step: bool,
        time_series_type: Literal["numeric", "string"],
        name: str | None | Omitted = OMITTED,
        description: str | None | Omitted = OMITTED,
        tags: list[str] | None | Omitted = OMITTED,
        aliases: list[str] | None | Omitted = OMITTED,
        source_id: str | None | Omitted = OMITTED,
        source_context: str | None | Omitted = OMITTED,
        source: DirectRelationReference | tuple[str, str] | None | Omitted = OMITTED,
        source_created_time: datetime | None | Omitted = OMITTED,
        source_updated_time: datetime | None | Omitted = OMITTED,
        source_created_user: str | None | Omitted = OMITTED,
        source_updated_user: str | None | Omitted = OMITTED,
        source_unit: str | None | Omitted = OMITTED,
        unit: DirectRelationReference | tuple[str, str] | None | Omitted = OMITTED,
        assets: list[DirectRelationReference | tuple[str, str]] | None | Omitted = OMITTED,
        equipment: list[DirectRelationReference | tuple[str, str]] | None | Omitted = OMITTED,
        extracted_data: dict | None | Omitted = OMITTED,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None | Omitted = OMITTED,
    ) -> None:
        TypedNodeApply.__init__(self, space, external_id, existing_version, type)
        self.is_step = is_step
        self.time_series_type = time_series_type
        self.name = name
        self.description = description
        self.tags = tags
        self.aliases = aliases
        self.source_id = source_id
        self.source_context = source_context
        self.source = DirectRelationReference.load(source) if source else source
        self.source_created_time = source_created_time
        self.source_updated_time = source_updated_time
        self.source_created_user = source_created_user
        self.source_updated_user = source_updated_user
        self.source_unit = source_unit
        self.unit = DirectRelationReference.load(unit) if unit else unit
        self.assets = [DirectRelationReference.load(a) for a in assets] if assets else assets
        self.equipment = [DirectRelationReference.load(e) for e in equipment] if equipment else equipment
        self.extracted_data = extracted_data


class CogniteExtractorTimeSeries(_CogniteExtractorTimeSeriesProperties, TypedNode):
    """This represents the reading format of Cognite extractor time series.

    It is used when data is read from CDF.

    Args:
        space (str): The space where the node is located.
        external_id (str): The external id of the Cognite extractor time series.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        is_step (bool): Defines whether the time series is a step series or not.
        time_series_type (Literal['numeric', 'string']): Defines data type of the data points.
        name (str | None): Name of the instance
        description (str | None): Description of the instance
        tags (list[str] | None): Text based labels for generic use, limited to 1000
        aliases (list[str] | None): Alternative names for the node
        source_id (str | None): Identifier from the source system
        source_context (str | None): Context of the source id. For systems where the sourceId is globally unique, the sourceContext is expected to not be set.
        source (DirectRelationReference | None): Direct relation to a source system
        source_created_time (datetime | None): When the instance was created in source system (if available)
        source_updated_time (datetime | None): When the instance was last updated in the source system (if available)
        source_created_user (str | None): User identifier from the source system on who created the source data. This identifier is not guaranteed to match the user identifiers in CDF
        source_updated_user (str | None): User identifier from the source system on who last updated the source data. This identifier is not guaranteed to match the user identifiers in CDF
        source_unit (str | None): Unit as specified in the source system
        unit (DirectRelationReference | None): direct relation to the unit of the time series
        assets (list[DirectRelationReference] | None): The asset field.
        equipment (list[DirectRelationReference] | None): The equipment field.
        extracted_data (dict | None): Unstructured information extracted from source system
        type (DirectRelationReference | None): Direct relation pointing to the type node.
        deleted_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        is_step: bool,
        time_series_type: Literal["numeric", "string"],
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        source_id: str | None = None,
        source_context: str | None = None,
        source: DirectRelationReference | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        source_unit: str | None = None,
        unit: DirectRelationReference | None = None,
        assets: list[DirectRelationReference] | None = None,
        equipment: list[DirectRelationReference] | None = None,
        extracted_data: dict | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.is_step = is_step
        self.time_series_type = time_series_type
        self.name = name
        self.description = description
        self.tags = tags
        self.aliases = aliases
        self.source_id = source_id
        self.source_context = source_context
        self.source = DirectRelationReference.load(source) if source else source
        self.source_created_time = source_created_time
        self.source_updated_time = source_updated_time
        self.source_created_user = source_created_user
        self.source_updated_user = source_updated_user
        self.source_unit = source_unit
        self.unit = DirectRelationReference.load(unit) if unit else unit
        self.assets = [DirectRelationReference.load(a) for a in assets] if assets else assets
        self.equipment = [DirectRelationReference.load(e) for e in equipment] if equipment else equipment
        self.extracted_data = extracted_data

    def as_write(self) -> CogniteExtractorTimeSeriesApply:
        return CogniteExtractorTimeSeriesApply(
            self.space,
            self.external_id,
            is_step=self.is_step,
            time_series_type=self.time_series_type,
            name=self.name,
            description=self.description,
            tags=self.tags,
            aliases=self.aliases,
            source_id=self.source_id,
            source_context=self.source_context,
            source=self.source,
            source_created_time=self.source_created_time,
            source_updated_time=self.source_updated_time,
            source_created_user=self.source_created_user,
            source_updated_user=self.source_updated_user,
            source_unit=self.source_unit,
            unit=self.unit,
            assets=self.assets,  # type: ignore[arg-type]
            equipment=self.equipment,  # type: ignore[arg-type]
            extracted_data=self.extracted_data,
            existing_version=self.version,
            type=self.type,
        )



from __future__ import annotations

from cognite.client._api.functions.utils import _get_function_identifier, _get_function_internal_id
from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes import (
    FunctionCall,
    FunctionCallList,
    FunctionCallLog,
)
from cognite.client.data_classes.functions import FunctionCallsFilter
from cognite.client.utils._identifier import IdentifierSequence


class FunctionCallsAPI(APIClient):
    _RESOURCE_PATH = "/functions/{}/calls"
    _RESOURCE_PATH_RESPONSE = "/functions/{}/calls/{}/response"
    _RESOURCE_PATH_LOGS = "/functions/{}/calls/{}/logs"

    async def list(
        self,
        function_id: int | None = None,
        function_external_id: str | None = None,
        status: str | None = None,
        schedule_id: int | None = None,
        start_time: dict[str, int] | None = None,
        end_time: dict[str, int] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
    ) -> FunctionCallList:
        """`List all calls associated with a specific function id. <https://developer.cognite.com/api#tag/Function-calls/operation/listFunctionCalls>`_ Either function_id or function_external_id must be specified.

        Args:
            function_id (int | None): ID of the function on which the calls were made.
            function_external_id (str | None): External ID of the function on which the calls were made.
            status (str | None): Status of the call. Possible values ["Running", "Failed", "Completed", "Timeout"].
            schedule_id (int | None): Schedule id from which the call belongs (if any).
            start_time (dict[str, int] | None): Start time of the call. Possible keys are `min` and `max`, with values given as time stamps in ms.
            end_time (dict[str, int] | None): End time of the call. Possible keys are `min` and `max`, with values given as time stamps in ms.
            limit (int | None): Maximum number of function calls to list. Pass in -1, float('inf') or None to list all Function Calls.

        Returns:
            FunctionCallList: List of function calls

        Examples:

            List function calls:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> calls = client.functions.calls.list(function_id=1)

            List function calls directly on a function object:

                >>> func = client.functions.retrieve(id=1)
                >>> calls = func.list_calls()

        """
        identifier = _get_function_identifier(function_id, function_external_id)
        function_id = await _get_function_internal_id(self._cognite_client, identifier)
        filter = FunctionCallsFilter(
            status=status,
            schedule_id=schedule_id,
            start_time=start_time,
            end_time=end_time,
        ).dump(camel_case=True)
        resource_path = self._RESOURCE_PATH.format(function_id)
        return await self._list(
            method="POST",
            resource_path=resource_path,
            filter=filter,
            limit=limit,
            resource_cls=FunctionCall,
            list_cls=FunctionCallList,
        )

    async def retrieve(
        self,
        call_id: int,
        function_id: int | None = None,
        function_external_id: str | None = None,
    ) -> FunctionCall | None:
        """`Retrieve a single function call by id. <https://developer.cognite.com/api#tag/Function-calls/operation/byIdsFunctionCalls>`_

        Args:
            call_id (int): ID of the call.
            function_id (int | None): ID of the function on which the call was made.
            function_external_id (str | None): External ID of the function on which the call was made.

        Returns:
            FunctionCall | None: Requested function call or None if either call ID or function identifier is not found.

        Examples:

            Retrieve single function call by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> call = client.functions.calls.retrieve(call_id=2, function_id=1)

            Retrieve function call directly on a function object:

                >>> func = client.functions.retrieve(id=1)
                >>> call = func.retrieve_call(id=2)
        """
        identifier = _get_function_identifier(function_id, function_external_id)
        function_id = await _get_function_internal_id(self._cognite_client, identifier)

        resource_path = self._RESOURCE_PATH.format(function_id)

        return await self._retrieve_multiple(
            resource_path=resource_path,
            identifiers=IdentifierSequence.load(ids=call_id).as_singleton(),
            resource_cls=FunctionCall,
            list_cls=FunctionCallList,
        )

    async def get_response(
        self,
        call_id: int,
        function_id: int | None = None,
        function_external_id: str | None = None,
    ) -> dict[str, object] | None:
        """`Retrieve the response from a function call. <https://developer.cognite.com/api#tag/Function-calls/operation/getFunctionCallResponse>`_

        Args:
            call_id (int): ID of the call.
            function_id (int | None): ID of the function on which the call was made.
            function_external_id (str | None): External ID of the function on which the call was made.

        Returns:
            dict[str, object] | None: Response from the function call.

        Examples:

            Retrieve function call response by call ID:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> response = client.functions.calls.get_response(call_id=2, function_id=1)

            Retrieve function call response directly on a call object:

                >>> call = client.functions.calls.retrieve(call_id=2, function_id=1)
                >>> response = call.get_response()

        """
        identifier = _get_function_identifier(function_id, function_external_id)
        function_id = await _get_function_internal_id(self._cognite_client, identifier)

        resource_path = self._RESOURCE_PATH_RESPONSE.format(function_id, call_id)
        response = await self._get(resource_path)
        return response.json().get("response")

    async def get_logs(
        self,
        call_id: int,
        function_id: int | None = None,
        function_external_id: str | None = None,
    ) -> FunctionCallLog:
        """`Retrieve logs for function call. <https://developer.cognite.com/api#tag/Function-calls/operation/getFunctionCalls>`_

        Args:
            call_id (int): ID of the call.
            function_id (int | None): ID of the function on which the call was made.
            function_external_id (str | None): External ID of the function on which the call was made.

        Returns:
            FunctionCallLog: Log for the function call.

        Examples:

            Retrieve function call logs by call ID:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> logs = client.functions.calls.get_logs(call_id=2, function_id=1)

            Retrieve function call logs directly on a call object:

                >>> call = client.functions.calls.retrieve(call_id=2, function_id=1)
                >>> logs = call.get_logs()

        """
        identifier = _get_function_identifier(function_id, function_external_id)
        function_id = await _get_function_internal_id(self._cognite_client, identifier)

        resource_path = self._RESOURCE_PATH_LOGS.format(function_id, call_id)
        response = await self._get(resource_path)
        return FunctionCallLog._load(response.json()["items"])



"""
===============================================================================
66e3c7c6a5b970455f6ecbbc6efaa5d2
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from cognite.client import AsyncCogniteClient
from cognite.client._sync_api.three_d.asset_mapping import Sync3DAssetMappingAPI
from cognite.client._sync_api.three_d.files import Sync3DFilesAPI
from cognite.client._sync_api.three_d.models import Sync3DModelsAPI
from cognite.client._sync_api.three_d.revisions import Sync3DRevisionsAPI
from cognite.client._sync_api_client import SyncAPIClient

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class Sync3DAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client
        self.models = Sync3DModelsAPI(async_client)
        self.revisions = Sync3DRevisionsAPI(async_client)
        self.files = Sync3DFilesAPI(async_client)
        self.asset_mappings = Sync3DAssetMappingAPI(async_client)



import typing
from typing import Literal

from cognite.client.utils._text import to_snake_case

Aggregate = Literal[
    "average",
    "continuous_variance",
    "count",
    "count_bad",
    "count_good",
    "count_uncertain",
    "discrete_variance",
    "duration_bad",
    "duration_good",
    "duration_uncertain",
    "interpolation",
    "max",
    "max_datapoint",
    "min",
    "min_datapoint",
    "step_interpolation",
    "sum",
    "total_variation",
]
_OBJECT_AGGREGATES_CAMEL: frozenset[Literal["maxDatapoint", "minDatapoint"]] = frozenset(
    {"maxDatapoint", "minDatapoint"}
)
OBJECT_AGGREGATES: frozenset[Literal["max_datapoint", "min_datapoint"]] = frozenset({"max_datapoint", "min_datapoint"})

# Assumption: All INT aggregates should adhere to the following logic: Missing values can be replace with 0.
#             Thus, if you add a new aggregate here, and this is no longer the case, a refactor is needed:
_INT_AGGREGATES_CAMEL = frozenset(
    {
        "count",
        "countBad",
        "countGood",
        "countUncertain",
        "durationBad",
        "durationGood",
        "durationUncertain",
    }
)
INT_AGGREGATES = frozenset(map(to_snake_case, _INT_AGGREGATES_CAMEL))
ALL_SORTED_DP_AGGS = sorted(typing.get_args(Aggregate))
_ALL_AGGREGATES = frozenset(ALL_SORTED_DP_AGGS)
ALL_SORTED_NUMERIC_DP_AGGS = [agg for agg in ALL_SORTED_DP_AGGS if agg not in ("min_datapoint", "max_datapoint")]

# When we add unit info to dataframe columns, we need to know if the physical unit should be included or not.
# Some, like variance, is techinally the unit squared, but we still include them (as opposed to count):
_AGGREGATES_WITH_UNIT = frozenset(
    {
        "average",
        "continuous_variance",
        "discrete_variance",
        "interpolation",
        "max",
        "max_datapoint",
        "min",
        "min_datapoint",
        "step_interpolation",
        "sum",
        "total_variation",
    }
)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from functools import cache
from typing import TYPE_CHECKING, Any, ClassVar, Literal, NoReturn, cast, get_args, get_type_hints

from typing_extensions import Self

from cognite.client.data_classes._base import CogniteResource, CogniteResourceList
from cognite.client.data_classes.data_modeling.ids import ContainerId
from cognite.client.data_classes.data_modeling.instances import InstanceSort
from cognite.client.utils._auxiliary import all_concrete_subclasses

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


@dataclass
class DebugInfo(CogniteResource):
    """
    Contains the requested debug information.

    Args:
        notices (DebugNoticeList | None): A list of notices that provide insights into the query's execution.
            These can highlight potential performance issues, offer optimization suggestions, or explain aspects
            of the query processing. Each notice falls into a category, such as indexing, sorting, filtering, or
            cursoring, to help identify areas for improvement.
        translated_query (TranslatedQuery | None): The internal representation of the query.
        plan (ExecutionPlan | None): The execution plan for the query.
    """

    notices: DebugNoticeList | None = None
    translated_query: TranslatedQuery | None = None
    plan: ExecutionPlan | None = None

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> DebugInfo:
        return cls(
            notices=DebugNoticeList._load(data["notices"], cognite_client) if "notices" in data else None,
            translated_query=TranslatedQuery._load(data["translatedQuery"]) if "translatedQuery" in data else None,
            plan=ExecutionPlan._load(data["plan"]) if "plan" in data else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj: dict[str, Any] = {}
        if self.notices is not None:
            obj["notices"] = self.notices.dump(camel_case=camel_case)
        if self.translated_query is not None:
            key = "translatedQuery" if camel_case else "translated_query"
            obj[key] = self.translated_query.dump(camel_case=camel_case)
        if self.plan is not None:
            obj["plan"] = self.plan.dump(camel_case=camel_case)
        return obj


@dataclass
class TranslatedQuery(CogniteResource):
    """
    Internal representation of query. Depends on postgres-controlled output, hence the generic dict types.

    Args:
        query (object): Parameterized query.
        parameters (object): Parameter values for query.
    """

    query: dict[str, Any]
    parameters: dict[str, Any]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> TranslatedQuery:
        return cls(query=data["query"], parameters=data["parameters"])

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {"query": self.query, "parameters": self.parameters}


@dataclass
class ExecutionPlan(CogniteResource):
    """
    Execution plan for the query.

    Args:
        full_plan (dict[str, Any]): The full execution plan.
        profiled (bool): The execution plan has been profiled.
        by_identifier (dict[str, Any]): The execution plan grouped by query identifiers.
    """

    full_plan: dict[str, Any]
    profiled: bool
    by_identifier: dict[str, Any]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> ExecutionPlan:
        return cls(
            full_plan=data["fullPlan"],
            profiled=data["profiled"],
            by_identifier=data["byIdentifier"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "fullPlan" if camel_case else "full_plan": self.full_plan,
            "profiled": self.profiled,
            "byIdentifier" if camel_case else "by_identifier": self.by_identifier,
        }


@dataclass(kw_only=True)
class DebugParameters:
    """
    Debug parameters for debugging and analyzing queries.

    Args:
        emit_results (bool): Include the query result in the response. Using emit_results=False is required for advanced query analysis features.
        timeout (int | None): Query timeout in milliseconds. Can be used to override the default timeout when analysing queries. Requires emit_results=False.
        include_translated_query (bool): Include the internal representation of the query.
        include_plan (bool): Include the execution plan for the query.
        profile (bool): Most thorough level of query analysis. Requires emit_results=False.
    """

    emit_results: bool = True
    timeout: int | None = None
    include_translated_query: bool = False
    include_plan: bool = False
    profile: bool = False

    @property
    def requires_alpha_header(self) -> bool:
        return self.include_translated_query or self.include_plan

    def dump(self, camel_case: bool = True) -> dict[str, bool | int]:
        res: dict[str, bool | int] = {
            "emitResults" if camel_case else "emit_results": self.emit_results,
            "profile": self.profile,
        }
        if self.timeout is not None:
            res["timeout"] = self.timeout
        if self.include_translated_query:
            key = "includeTranslatedQuery" if camel_case else "include_translated_query"
            res[key] = self.include_translated_query
        if self.include_plan:
            key = "includePlan" if camel_case else "include_plan"
            res[key] = self.include_plan
        return res


@dataclass
class DebugNotice(CogniteResource, ABC):
    """
    A notice that provides insight into the query's execution. It can highlight potential performance issues,
    offer an optimization suggestion, or explain an aspect of the query processing. Each notice falls into a
    category, such as indexing, sorting, filtering, or cursoring, to help identify areas for improvement.
    """

    category: str
    code: str
    hint: str
    level: str

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> DebugNotice:
        subclass_lookup = _create_debug_notice_subclass_map()
        try:
            key = (data["code"], data["category"])
            return subclass_lookup[key]._load(data)
        except KeyError:
            return cast(DebugNotice, UnknownDebugNotice._load(data))

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "category": self.category,
            "code": self.code,
            "hint": self.hint,
            "level": self.level,
        }


@dataclass
class InvalidDebugOptionsNotice(DebugNotice, ABC):
    pass


@dataclass
class ExcessiveTimeoutNotice(InvalidDebugOptionsNotice):
    code: Literal["excessiveTimeout"]
    category: Literal["invalidDebugOptions"]
    level: Literal["warning"]
    timeout: int

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            hint=data["hint"],
            timeout=data["timeout"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj = super().dump(camel_case=camel_case)
        obj["timeout"] = self.timeout
        return obj


@dataclass
class NoTimeoutWithResultsNotice(InvalidDebugOptionsNotice):
    code: Literal["noTimeoutWithResults"]
    category: Literal["invalidDebugOptions"]
    level: Literal["warning"]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            hint=data["hint"],
        )


@dataclass
class CursoringNotice(DebugNotice, ABC):
    pass


@dataclass
class IntractableDirectRelationsCursorNotice(CursoringNotice):
    code: Literal["intractableDirectRelationsCursor"]
    category: Literal["cursoring"]
    level: Literal["warning"]
    grade: Literal["D"]
    result_expression: str

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            grade=data["grade"],
            hint=data["hint"],
            result_expression=data["resultExpression"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj = super().dump(camel_case=camel_case)
        obj["grade"] = self.grade
        obj["resultExpression" if camel_case else "result_expression"] = self.result_expression
        return obj


@dataclass
class IndexingNotice(DebugNotice, ABC):
    pass


@dataclass
class UnindexedThroughNotice(IndexingNotice):
    code: Literal["unindexedThrough"]
    category: Literal["indexing"]
    level: Literal["warning"]
    grade: Literal["E"]
    result_expression: str
    property: list[str]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            grade=data["grade"],
            hint=data["hint"],
            result_expression=data["resultExpression"],
            property=data["property"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj = super().dump(camel_case=camel_case)
        obj.update(
            {
                "resultExpression" if camel_case else "result_expression": self.result_expression,
                "grade": self.grade,
                "property": self.property,
            }
        )
        return obj


@dataclass
class ContainersWithoutIndexesInvolvedNotice(IndexingNotice):
    code: Literal["containersWithoutIndexesInvolved"]
    category: Literal["indexing"]
    level: Literal["warning"]
    grade: Literal["C"]
    result_expression: str | None
    containers: list[ContainerId]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            grade=data["grade"],
            hint=data["hint"],
            result_expression=data.get("resultExpression"),
            containers=[ContainerId.load(c) for c in data["containers"]],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj = super().dump(camel_case=camel_case)
        obj.update(
            {
                "grade": self.grade,
                "containers": [c.dump(camel_case=camel_case) for c in self.containers],
                "resultExpression" if camel_case else "result_expression": self.result_expression,
            }
        )
        return obj


@dataclass
class FilteringNotice(DebugNotice, ABC):
    pass


@dataclass
class SelectiveExternalIDFilterNotice(FilteringNotice):
    code: Literal["selectiveExternalIDFilter"]
    category: Literal["filtering"]
    level: Literal["info"]
    grade: Literal["A"]
    result_expression: str
    via_from: str | None

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            grade=data["grade"],
            hint=data["hint"],
            result_expression=data["resultExpression"],
            via_from=data.get("viaFrom"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj = super().dump(camel_case=camel_case)
        obj.update(
            {
                "resultExpression" if camel_case else "result_expression": self.result_expression,
                "viaFrom" if camel_case else "via_from": self.via_from,
                "grade": self.grade,
            }
        )
        return obj


@dataclass
class SignificantHasDataFiltersNotice(FilteringNotice):
    code: Literal["significantHasDataFiltering"]
    category: Literal["filtering"]
    level: Literal["warning"]
    grade: Literal["C"]
    result_expression: str
    containers: list[ContainerId]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            grade=data["grade"],
            hint=data["hint"],
            result_expression=data["resultExpression"],
            containers=[ContainerId.load(c) for c in data["containers"]],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj = super().dump(camel_case=camel_case)
        obj.update(
            {
                "grade": self.grade,
                "resultExpression" if camel_case else "result_expression": self.result_expression,
                "containers": [c.dump(camel_case=camel_case) for c in self.containers],
            }
        )
        return obj


@dataclass
class SignificantPostFilteringNotice(FilteringNotice):
    code: Literal["significantPostFiltering"]
    category: Literal["filtering"]
    level: Literal["warning"]
    grade: Literal["C"]
    result_expression: str
    limit: int
    max_involved_rows: int

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            grade=data["grade"],
            hint=data["hint"],
            result_expression=data["resultExpression"],
            limit=data["limit"],
            max_involved_rows=data["maxInvolvedRows"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj = super().dump(camel_case=camel_case)
        obj.update(
            {
                "grade": self.grade,
                "resultExpression" if camel_case else "result_expression": self.result_expression,
                "limit": self.limit,
                "maxInvolvedRows" if camel_case else "max_involved_rows": self.max_involved_rows,
            }
        )
        return obj


@dataclass
class SortingNotice(DebugNotice, ABC):
    pass


@dataclass
class SortNotBackedByIndexNotice(SortingNotice):
    code: Literal["sortNotBackedByIndex"]
    category: Literal["sorting"]
    level: Literal["warning"]
    grade: Literal["C"]
    result_expression: str
    sort: list[InstanceSort]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            grade=data["grade"],
            hint=data["hint"],
            result_expression=data["resultExpression"],
            sort=[InstanceSort.load(s) for s in data["sort"]],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj = super().dump(camel_case=camel_case)
        obj.update(
            {
                "grade": self.grade,
                "resultExpression" if camel_case else "result_expression": self.result_expression,
                "sort": [s.dump(camel_case=camel_case) for s in self.sort],
            }
        )
        return obj


@dataclass
class FilterMatchesCursorableSortNotice(SortingNotice):
    code: Literal["filterMatchesCursorableSort"]
    category: Literal["sorting"]
    level: Literal["info"]
    grade: Literal["A", "B"]
    result_expression: str
    sort: list[InstanceSort]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            code=data["code"],
            category=data["category"],
            level=data["level"],
            grade=data["grade"],
            hint=data["hint"],
            result_expression=data["resultExpression"],
            sort=[InstanceSort.load(s) for s in data["sort"]],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        obj = super().dump(camel_case=camel_case)
        obj.update(
            {
                "grade": self.grade,
                "resultExpression" if camel_case else "result_expression": self.result_expression,
                "sort": [s.dump(camel_case=camel_case) for s in self.sort],
            }
        )
        return obj


@dataclass
class UnknownDebugNotice(DebugNotice):
    _MISSING: ClassVar[str] = "MISSING/UNKNOWN"
    data: dict[str, Any]

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        data = data.copy()
        category = data.pop("category", cls._MISSING)
        code = data.pop("code", cls._MISSING)
        hint = data.pop("hint", cls._MISSING)
        level = data.pop("level", cls._MISSING)
        return cls(category=category, code=code, hint=hint, level=level, data=data)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        if camel_case is False:
            import warnings

            warnings.warn(
                f"{type(self).__name__}.dump(...) does not support 'camel_case=False', returning original data. "
                "Please raise an issue on Github to add support for this debug notice type.",
                stacklevel=2,
            )
        output = {}
        for field in ["category", "code", "hint", "level"]:
            value = getattr(self, field)
            if value is not None and value != self._MISSING:
                output[field] = value
        return output | self.data.copy()


class DebugNoticeList(CogniteResourceList[DebugNotice]):
    _RESOURCE = DebugNotice

    def get(self, *a: Any, **kw: Any) -> NoReturn:
        raise NotImplementedError


@cache
def _create_debug_notice_subclass_map() -> dict[tuple[str, str], type[DebugNotice]]:
    from cognite.client import AsyncCogniteClient, CogniteClient

    additional_refs = {"CogniteClient": CogniteClient, "AsyncCogniteClient": AsyncCogniteClient}

    lookup = {}
    subclasses: set[type[DebugNotice]] = all_concrete_subclasses(DebugNotice, exclude={UnknownDebugNotice})
    for sub_cls in subclasses:
        # When we do 'get_type_hints', we evaluate forward refs., and we reference InstanceSort in a file
        # that only imports CogniteClient while TYPE_CHECKING, so it is not available at runtime which we need:
        annots = get_type_hints(sub_cls, globalns=globals() | additional_refs)
        code = get_args(annots["code"])[0]
        category = get_args(annots["category"])[0]
        lookup[code, category] = sub_cls
    return lookup



from __future__ import annotations

from typing import overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.user_profiles import UserProfile, UserProfileList, UserProfilesConfiguration
from cognite.client.utils._identifier import UserIdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr


class UserProfilesAPI(APIClient):
    _RESOURCE_PATH = "/profiles"

    async def enable(self) -> UserProfilesConfiguration:
        """Enable user profiles for the project"""
        res = await self._post("/update", json={"update": {"userProfilesConfiguration": {"set": {"enabled": True}}}})
        return UserProfilesConfiguration._load(res.json()["userProfilesConfiguration"])

    async def disable(self) -> UserProfilesConfiguration:
        """Disable user profiles for the project"""
        res = await self._post("/update", json={"update": {"userProfilesConfiguration": {"set": {"enabled": False}}}})
        return UserProfilesConfiguration._load(res.json()["userProfilesConfiguration"])

    async def me(self) -> UserProfile:
        """`Retrieve your own user profile <https://developer.cognite.com/api#tag/User-profiles/operation/getRequesterUserProfile>`_

        Retrieves the user profile of the principal issuing the request, i.e. the principal *this* AsyncCogniteClient was instantiated with.

        Returns:
            UserProfile: Your own user profile.

        Raises:
            CogniteAPIError: If this principal doesn't have a user profile, you get a not found (404) response code.

        Examples:

            Get your own user profile:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.user_profiles.me()
        """
        return UserProfile._load((await self._get(self._RESOURCE_PATH + "/me")).json())

    @overload
    async def retrieve(self, user_identifier: str) -> UserProfile | None: ...

    @overload
    async def retrieve(self, user_identifier: SequenceNotStr[str]) -> UserProfileList: ...

    async def retrieve(self, user_identifier: str | SequenceNotStr[str]) -> UserProfile | UserProfileList | None:
        """`Retrieve user profiles by user identifier. <https://developer.cognite.com/api#tag/User-profiles/operation/getUserProfilesByIds>`_

        Retrieves one or more user profiles indexed by the user identifier in the same CDF project.

        Args:
            user_identifier (str | SequenceNotStr[str]): The single user identifier (or sequence of) to retrieve profile(s) for.

        Returns:
            UserProfile | UserProfileList | None: UserProfileList if a sequence of user identifier were requested, else UserProfile. If a single user identifier is requested and it is not found, None is returned.

        Raises:
            CogniteNotFoundError: A sequences of user identifiers were requested, but one or more does not exist.

        Examples:

            Get a single user profile:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.user_profiles.retrieve("foo")

            Get multiple user profiles:

                >>> res = client.iam.user_profiles.retrieve(["bar", "baz"])
        """
        identifiers = UserIdentifierSequence.load(user_identifier)
        return await self._retrieve_multiple(
            list_cls=UserProfileList,
            resource_cls=UserProfile,
            identifiers=identifiers,
        )

    async def search(self, name: str, limit: int = DEFAULT_LIMIT_READ) -> UserProfileList:
        """`Search for user profiles <https://developer.cognite.com/api#tag/User-profiles/operation/userProfilesSearch>`_
        Primarily meant for human-centric use-cases and data exploration, not for programs, as the result set ordering and match criteria threshold may change over time.

        Args:
            name (str): Prefix search on name.
            limit (int): Maximum number of results to return.

        Returns:
            UserProfileList: User profiles search result

        Examples:

            Search for users with first (or second...) name starting with "Alex":

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.user_profiles.search(name="Alex")
        """
        return await self._search(
            list_cls=UserProfileList,
            search={"name": name},
            filter={},
            limit=limit,
        )

    async def list(self, limit: int | None = DEFAULT_LIMIT_READ) -> UserProfileList:
        """`List user profiles <https://developer.cognite.com/api#tag/User-profiles/operation/listUserProfiles>`_

        List all user profiles in the current CDF project. The results are ordered alphabetically by name.

        Args:
            limit (int | None): Maximum number of user profiles to return. Defaults to 25. Set to -1, float("inf") or None to return all.

        Returns:
            UserProfileList: List of user profiles.

        Examples:

            List all user profiles:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.user_profiles.list(limit=None)
        """
        return await self._list(
            "GET",
            list_cls=UserProfileList,
            resource_cls=UserProfile,
            limit=limit,
        )



from __future__ import annotations

from cognite.client.data_classes import aggregations, filters
from cognite.client.data_classes.aggregations import AggregatedValue, Aggregation
from cognite.client.data_classes.data_modeling import query
from cognite.client.data_classes.data_modeling.containers import (
    BTreeIndex,
    BTreeIndexApply,
    Constraint,
    ConstraintApply,
    Container,
    ContainerApply,
    ContainerApplyList,
    ContainerList,
    ContainerProperty,
    ContainerPropertyApply,
    Index,
    IndexApply,
    InvertedIndex,
    InvertedIndexApply,
    RequiresConstraint,
    RequiresConstraintApply,
    UniquenessConstraint,
    UniquenessConstraintApply,
)
from cognite.client.data_classes.data_modeling.data_models import (
    DataModel,
    DataModelApply,
    DataModelApplyList,
    DataModelFilter,
    DataModelList,
    DataModelsSort,
)
from cognite.client.data_classes.data_modeling.data_types import (
    Boolean,
    CDFExternalIdReference,
    Date,
    DirectRelation,
    DirectRelationReference,
    FileReference,
    Float32,
    Float64,
    Int32,
    Int64,
    Json,
    PropertyType,
    SequenceReference,
    Text,
    TimeSeriesReference,
    Timestamp,
)
from cognite.client.data_classes.data_modeling.debug import (
    DebugNotice,
    DebugNoticeList,
    DebugParameters,
    ExecutionPlan,
    TranslatedQuery,
)
from cognite.client.data_classes.data_modeling.ids import (
    ContainerId,
    ContainerIdentifier,
    DataModelId,
    DataModelIdentifier,
    DataModelingId,
    EdgeId,
    NodeId,
    PropertyId,
    VersionedDataModelingId,
    ViewId,
    ViewIdentifier,
)
from cognite.client.data_classes.data_modeling.instances import (
    Edge,
    EdgeApply,
    EdgeApplyList,
    EdgeApplyResult,
    EdgeApplyResultList,
    EdgeList,
    EdgeListWithCursor,
    InstanceApply,
    InstancesApply,
    InstancesApplyResult,
    InstancesDeleteResult,
    InstanceSort,
    InstancesResult,
    InvolvedContainers,
    InvolvedViews,
    Node,
    NodeApply,
    NodeApplyList,
    NodeApplyResult,
    NodeApplyResultList,
    NodeList,
    NodeListWithCursor,
    NodeOrEdgeData,
    PropertyOptions,
    TypedEdge,
    TypedEdgeApply,
    TypedNode,
    TypedNodeApply,
)
from cognite.client.data_classes.data_modeling.query import (
    EdgeResultSetExpression,
    Intersection,
    NodeOrEdgeResultSetExpression,
    NodeResultSetExpression,
    Query,
    QueryResult,
    ResultSetExpression,
    Select,
    SetOperation,
    SourceSelector,
    Union,
    UnionAll,
)
from cognite.client.data_classes.data_modeling.spaces import Space, SpaceApply, SpaceApplyList, SpaceList
from cognite.client.data_classes.data_modeling.sync import SubscriptionContext
from cognite.client.data_classes.data_modeling.views import (
    ConnectionDefinition,
    EdgeConnection,
    EdgeConnectionApply,
    MappedProperty,
    MappedPropertyApply,
    MultiEdgeConnection,
    MultiEdgeConnectionApply,
    MultiReverseDirectRelation,
    MultiReverseDirectRelationApply,
    SingleHopConnectionDefinition,
    View,
    ViewApply,
    ViewApplyList,
    ViewFilter,
    ViewList,
)
from cognite.client.data_classes.filters import Filter

__all__ = [
    "AggregatedValue",
    "Aggregation",
    "BTreeIndex",
    "BTreeIndexApply",
    "Boolean",
    "CDFExternalIdReference",
    "ConnectionDefinition",
    "Constraint",
    "ConstraintApply",
    "Container",
    "ContainerApply",
    "ContainerApplyList",
    "ContainerId",
    "ContainerIdentifier",
    "ContainerList",
    "ContainerProperty",
    "ContainerPropertyApply",
    "DataModel",
    "DataModelApply",
    "DataModelApplyList",
    "DataModelFilter",
    "DataModelId",
    "DataModelIdentifier",
    "DataModelList",
    "DataModelingId",
    "DataModelsSort",
    "Date",
    "DebugNotice",
    "DebugNoticeList",
    "DebugParameters",
    "DirectRelation",
    "DirectRelationReference",
    "Edge",
    "EdgeApply",
    "EdgeApplyList",
    "EdgeApplyResult",
    "EdgeApplyResultList",
    "EdgeConnection",
    "EdgeConnectionApply",
    "EdgeId",
    "EdgeList",
    "EdgeListWithCursor",
    "EdgeResultSetExpression",
    "ExecutionPlan",
    "FileReference",
    "Filter",
    "Float32",
    "Float64",
    "Index",
    "IndexApply",
    "InstanceApply",
    "InstanceSort",
    "InstancesApply",
    "InstancesApplyResult",
    "InstancesDeleteResult",
    "InstancesResult",
    "Int32",
    "Int64",
    "Intersection",
    "InvertedIndex",
    "InvertedIndexApply",
    "InvolvedContainers",
    "InvolvedViews",
    "Json",
    "MappedProperty",
    "MappedPropertyApply",
    "MultiEdgeConnection",
    "MultiEdgeConnectionApply",
    "MultiReverseDirectRelation",
    "MultiReverseDirectRelationApply",
    "Node",
    "NodeApply",
    "NodeApplyList",
    "NodeApplyResult",
    "NodeApplyResultList",
    "NodeId",
    "NodeList",
    "NodeListWithCursor",
    "NodeOrEdgeData",
    "NodeOrEdgeResultSetExpression",
    "NodeResultSetExpression",
    "PropertyId",
    "PropertyOptions",
    "PropertyType",
    "Query",
    "QueryResult",
    "RequiresConstraint",
    "RequiresConstraintApply",
    "ResultSetExpression",
    "Select",
    "SequenceReference",
    "SetOperation",
    "SingleHopConnectionDefinition",
    "SourceSelector",
    "Space",
    "SpaceApply",
    "SpaceApplyList",
    "SpaceList",
    "SubscriptionContext",
    "Text",
    "TimeSeriesReference",
    "Timestamp",
    "TranslatedQuery",
    "TypedEdge",
    "TypedEdgeApply",
    "TypedNode",
    "TypedNodeApply",
    "Union",
    "UnionAll",
    "UniquenessConstraint",
    "UniquenessConstraintApply",
    "VersionedDataModelingId",
    "View",
    "ViewApply",
    "ViewApplyList",
    "ViewFilter",
    "ViewId",
    "ViewIdentifier",
    "ViewList",
    "aggregations",
    "filters",
    "query",
]



"""
===============================================================================
5268479111509d912fc224eb231afa08
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from cognite.client import AsyncCogniteClient
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.iam import TokenInspection
from cognite.client.utils._async_helpers import run_sync


class SyncTokenAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    def inspect(self) -> TokenInspection:
        """
        Inspect a token.

        Get details about which projects it belongs to and which capabilities are granted to it.

        Returns:
            TokenInspection: The object with token inspection details.

        Example:

            Inspect token::

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.token.inspect()
        """
        return run_sync(self.__async_client.iam.token.inspect())



from __future__ import annotations

from typing import TYPE_CHECKING

from cognite.client._api.ai.tools.documents import AIDocumentsAPI
from cognite.client._api_client import APIClient

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient, ClientConfig


class AIToolsAPI(APIClient):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: AsyncCogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.documents = AIDocumentsAPI(config, api_version, cognite_client)



"""
===============================================================================
e53e36c6b07aeb024b15078c9e4ca307
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

from cognite.client import AsyncCogniteClient
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.ai import Answer, AnswerLanguage, Summary
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.utils._async_helpers import run_sync


class SyncAIDocumentsAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    def summarize(
        self, id: int | None = None, external_id: str | None = None, instance_id: NodeId | None = None
    ) -> Summary:
        """
        `Summarize a document using a Large Language Model. <https://developer.cognite.com/api#tag/Document-AI/operation/document_questioning_api_v1_projects__projectName__ai_tools_documents_ask_post>`_

        Note:
            Currently only supports summarizing a single document at a time, but
            this may be extended in the future.

        Args:
            id (int | None): The ID of the document
            external_id (str | None): The external ID of the document
            instance_id (NodeId | None): The instance ID of the document

        Returns:
            Summary: A summary of the document.

        Examples:

            Summarize a single document using ID:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> client.ai.tools.documents.summarize(id=123)

            You can also use external ID or instance ID:

                >>> from cognite.client.data_classes.data_modeling import NodeId
                >>> client.ai.tools.documents.summarize(
                ...     instance_id=NodeId("my-space", "my-xid")
                ... )
        """
        return run_sync(
            self.__async_client.ai.tools.documents.summarize(id=id, external_id=external_id, instance_id=instance_id)
        )

    def ask_question(
        self,
        question: str,
        *,
        id: int | Sequence[int] | None = None,
        external_id: str | Sequence[str] | None = None,
        instance_id: NodeId | Sequence[NodeId] | None = None,
        language: AnswerLanguage
        | Literal[
            "Chinese",
            "Dutch",
            "English",
            "French",
            "German",
            "Italian",
            "Japanese",
            "Korean",
            "Latvian",
            "Norwegian",
            "Portuguese",
            "Spanish",
            "Swedish",
        ] = AnswerLanguage.English,
        additional_context: str | None = None,
        ignore_unknown_ids: bool = False,
    ) -> Answer:
        """
        `Ask a question about one or more documents using a Large Language Model. <https://developer.cognite.com/api#tag/Document-AI/operation/documents_summary_api_v1_projects__projectName__ai_tools_documents_summarize_post>`_

        Supports up to 100 documents at a time.

        Args:
            question (str): The question.
            id (int | Sequence[int] | None): The ID(s) of the document(s)
            external_id (str | Sequence[str] | None): The external ID(s) of the document(s)
            instance_id (NodeId | Sequence[NodeId] | None): The instance ID(s) of the document(s)
            language (AnswerLanguage | Literal['Chinese', 'Dutch', 'English', 'French', 'German', 'Italian', 'Japanese', 'Korean', 'Latvian', 'Norwegian', 'Portuguese', 'Spanish', 'Swedish']): The desired language of the answer, defaults to English.
            additional_context (str | None): Additional context that you want the LLM to take into account.
            ignore_unknown_ids (bool): Whether to skip documents that do not exist or that are not fully processed, instead of throwing an error. If no valid documents are found, an error will always be raised.

        Returns:
            Answer: The answer to the question in the form of a list of multiple content objects, each consisting of a chunk of text along with a set of references.

        Examples:

            Ask a question about a single document with id=123 and get the answer in English (default):

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> client.ai.tools.documents.ask_question(
                ...     question="What model pump was used?",
                ...     id=123,
                ... )

            Ask a question about multiple documents referenced using external IDs, and instance ID
            and get the answer in German:

                >>> from cognite.client.data_classes.data_modeling import NodeId
                >>> from cognite.client.data_classes.ai import AnswerLanguage
                >>> client.ai.tools.documents.ask_question(
                ...     question="What other pumps are available?",
                ...     external_id=["foo", "bar"],
                ...     instance_id=NodeId("my-space", "my-xid"),
                ...     language=AnswerLanguage.German,
                ... )
        """
        return run_sync(
            self.__async_client.ai.tools.documents.ask_question(
                question=question,
                id=id,
                external_id=external_id,
                instance_id=instance_id,
                language=language,
                additional_context=additional_context,
                ignore_unknown_ids=ignore_unknown_ids,
            )
        )



from __future__ import annotations

from collections.abc import Sequence
from typing import overload
from urllib.parse import urljoin

from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._org_client import OrgAPIClient
from cognite.client.data_classes.principals import Principal, PrincipalList
from cognite.client.utils._identifier import PrincipalIdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr


class PrincipalsAPI(OrgAPIClient):
    _RESOURCE_PATH = "/principals"

    async def me(self) -> Principal:
        """`Get the current caller's information. <https://developer.cognite.com/api#tag/Principals/operation/getMe>`_

        Returns:
            Principal: The principal of the user running the code, i.e. the
                principal *this* AsyncCogniteClient was instantiated with.

        Examples:
            Get your own principal:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.principals.me()
        """
        # the /me endpoint is not using the /orgs/{org} base path, so we have to construct the URL manually
        path = f"/api/{self._api_version}{self._RESOURCE_PATH}/me"
        full_url = urljoin(self._AUTH_URL, path)

        response = await self._cognite_client._a.get(url=full_url)  # type: ignore [attr-defined]
        return Principal._load(response.json())

    @overload
    async def retrieve(
        self,
        id: str,
        external_id: None = None,
        ignore_unknown_ids: bool = False,
    ) -> Principal | None: ...

    @overload
    async def retrieve(
        self,
        id: None = None,
        *,
        external_id: str,
        ignore_unknown_ids: bool = False,
    ) -> Principal | None: ...

    @overload
    async def retrieve(
        self,
        id: SequenceNotStr[str],
        external_id: None = None,
        ignore_unknown_ids: bool = False,
    ) -> PrincipalList: ...

    @overload
    async def retrieve(
        self,
        id: None = None,
        *,
        external_id: SequenceNotStr[str],
        ignore_unknown_ids: bool = False,
    ) -> PrincipalList: ...

    @overload
    async def retrieve(
        self,
        id: None = None,
        external_id: None = None,
        ignore_unknown_ids: bool = False,
    ) -> PrincipalList: ...

    async def retrieve(
        self,
        id: str | SequenceNotStr[str] | None = None,
        external_id: str | SequenceNotStr[str] | None = None,
        ignore_unknown_ids: bool = False,
    ) -> Principal | PrincipalList | None:
        """`Retrieve principal by reference in the organization <https://developer.cognite.com/api#tag/Principals/operation/getPrincipalsById>`_

        Args:
            id (str | SequenceNotStr[str] | None): The ID(s) of the principal(s) to retrieve.
            external_id (str | SequenceNotStr[str] | None): The external ID(s) of the principal to retrieve.
            ignore_unknown_ids (bool): This is only relevant when retrieving multiple principals. If set to True, the method will return the principals that were found and ignore the ones that were not found. If set to False, the method will raise a CogniteAPIError if any of the specified principals were not found. Defaults to False.

        Returns:
            Principal | PrincipalList | None: The principal(s) with the specified ID(s) or external ID(s).

        Examples:
            Retrieve a principal by ID:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.principals.retrieve(id="20u3of8-1234-5678-90ab-cdef12345678")

            Retrieve a principal by external ID:
                >>> res = client.iam.principals.retrieve(external_id="my_external_id")
        """
        identifier = PrincipalIdentifierSequence.load(ids=id, external_ids=external_id)
        return await self._retrieve_multiple(
            list_cls=PrincipalList,
            resource_cls=Principal,  # type: ignore[type-abstract]
            identifiers=identifier,
            other_params={"ignoreUnknownIds": True if identifier.is_singleton() else ignore_unknown_ids},
        )

    async def list(self, types: str | Sequence[str] | None = None, limit: int = DEFAULT_LIMIT_READ) -> PrincipalList:
        """`List principals in the organization <https://developer.cognite.com/api#tag/Principals/operation/listPrincipals>`_

        Args:
            types (str | Sequence[str] | None): Filter by principal type(s). Defaults to None, which means no filtering.
            limit (int): The maximum number of principals to return. Defaults to 25.

        Returns:
            PrincipalList: The principal of the user running the code, i.e. the principal *this* CogniteClient was instantiated with.

        Examples:
            List principals in the organization:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.principals.list(types="USER", limit=10)
        """
        other_params: dict[str, str | list[str]] | None = None
        match types:
            case str():
                other_params = {"types": types.upper()}
            case Sequence():
                other_params = {"types": [t.upper() for t in types]}

        return await self._list(
            method="GET",
            list_cls=PrincipalList,
            # The Principal is abstract, but calling load on it will return a concrete instance:
            resource_cls=Principal,  # type: ignore[type-abstract]
            limit=limit,
            other_params=other_params,
        )



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CognitePrimitiveUpdate,
    CogniteResource,
    CogniteResourceList,
    CogniteUpdate,
    ExternalIDTransformerMixin,
    IdTransformerMixin,
    PropertySpec,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.utils._async_helpers import run_sync
from cognite.client.utils._text import copy_doc_from_async

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class SimulatorModelRevisionCore(WriteableCogniteResource["SimulatorModelRevisionWrite"], ABC):
    def __init__(
        self,
        external_id: str,
        model_external_id: str,
        file_id: int,
        description: str | None = None,
        external_dependencies: list[SimulatorModelRevisionDependency] | None = None,
    ) -> None:
        self.external_id = external_id
        self.model_external_id = model_external_id
        self.file_id = file_id
        self.description = description
        self.external_dependencies = external_dependencies

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.external_dependencies is not None:
            output["externalDependencies"] = [item.dump(camel_case=camel_case) for item in self.external_dependencies]

        return output


class SimulatorModelRevisionWrite(SimulatorModelRevisionCore):
    def as_write(self) -> SimulatorModelRevisionWrite:
        """Returns a writeable version of this resource"""
        return self

    @classmethod
    def _load(
        cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None
    ) -> SimulatorModelRevisionWrite:
        return cls(
            external_id=resource["externalId"],
            model_external_id=resource["modelExternalId"],
            file_id=resource["fileId"],
            description=resource.get("description"),
            external_dependencies=SimulatorModelRevisionDependency._load_list(
                resource["externalDependencies"], cognite_client
            )
            if "externalDependencies" in resource
            else None,
        )


class SimulatorModelRevision(SimulatorModelRevisionCore):
    """
    Simulator model revisions track changes and updates to a simulator model over time.

    Each revision ensures that modifications to models are traceable and allows users to understand the evolution of a given model.

    Args:
        id (int): Internal id of the simulator model revision
        external_id (str): External id of the simulator model revision
        model_external_id (str): External id of the associated simulator model
        file_id (int): The id of the file associated with the simulator model revision
        created_time (int): The time when the simulator model revision was created
        last_updated_time (int): The time when the simulator model revision was last updated
        simulator_external_id (str): External id of the simulator associated with the simulator model revision
        data_set_id (int): The id of the dataset associated with the simulator model revision
        created_by_user_id (str): The id of the user who created the simulator model revision
        status (str): The status of the simulator model revision
        version_number (int): The version number of the simulator model revision
        log_id (int): The id of the log associated with the simulator model revision
        description (str | None): The description of the simulator model revision
        status_message (str | None): The current status message of the simulator model revision
        external_dependencies (list[SimulatorModelRevisionDependency] | None): A list of external dependencies for the simulator model revision
        cognite_client (AsyncCogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        id: int,
        external_id: str,
        model_external_id: str,
        file_id: int,
        created_time: int,
        last_updated_time: int,
        simulator_external_id: str,
        data_set_id: int,
        created_by_user_id: str,
        status: str,
        version_number: int,
        log_id: int,
        description: str | None = None,
        status_message: str | None = None,
        external_dependencies: list[SimulatorModelRevisionDependency] | None = None,
        cognite_client: AsyncCogniteClient | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            model_external_id=model_external_id,
            file_id=file_id,
            description=description,
            external_dependencies=external_dependencies,
        )
        self.id = id
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.data_set_id = data_set_id
        self.created_by_user_id = created_by_user_id
        self.status = status
        self.version_number = version_number
        self.log_id = log_id
        self.status_message = status_message
        self.simulator_external_id = simulator_external_id
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            id=resource["id"],
            external_id=resource["externalId"],
            simulator_external_id=resource["simulatorExternalId"],
            model_external_id=resource["modelExternalId"],
            data_set_id=resource["dataSetId"],
            file_id=resource["fileId"],
            created_by_user_id=resource["createdByUserId"],
            status=resource["status"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            version_number=resource["versionNumber"],
            log_id=resource["logId"],
            description=resource.get("description"),
            status_message=resource.get("statusMessage"),
            external_dependencies=SimulatorModelRevisionDependency._load_list(
                resource["externalDependencies"], cognite_client
            )
            if "externalDependencies" in resource
            else None,
            cognite_client=cognite_client,
        )

    def as_write(self) -> SimulatorModelRevisionWrite:
        """Returns this SimulatorModelRevision in its writing version."""
        return SimulatorModelRevisionWrite(
            external_id=self.external_id,
            model_external_id=self.model_external_id,
            file_id=self.file_id,
            description=self.description,
        )

    async def get_data_async(self) -> SimulatorModelRevisionData | None:
        """`Retrieve data associated with this simulator model revision. <https://developer.cognite.com/api#tag/Simulator-Models/operation/retrieve_simulator_model_revision_data>`_

        Returns:
            SimulatorModelRevisionData | None: Data for the simulator model revision.
        """
        data = await self._cognite_client.simulators.models.revisions.retrieve_data(
            model_revision_external_id=self.external_id
        )
        if data:
            return data[0]
        return None

    @copy_doc_from_async(get_data_async)
    def get_data(self) -> SimulatorModelRevisionData | None:
        return run_sync(self.get_data_async())


class SimulatorModelCore(WriteableCogniteResource["SimulatorModelWrite"], ABC):
    """
    The simulator model resource represents an asset modeled in a simulator.

    This asset could range from a pump or well to a complete processing facility or refinery.
    The simulator model is the root of its associated revisions, routines, runs, and results.
    The dataset assigned to a model is inherited by its children. Deleting a model also deletes all its children, thereby
    maintaining the integrity and hierarchy of the simulation data.
    Simulator model revisions track changes and updates to a simulator model over time.
    Each revision ensures that modifications to models are traceable and allows users to understand the evolution of a given model.

    This is the read/response format of a simulator model.

    Args:
        external_id (str): External id of the simulator model
        simulator_external_id (str): External id of the associated simulator
        data_set_id (int): The id of the dataset associated with the simulator model
        name (str): The name of the simulator model
        type (str): The type key of the simulator model
        description (str | None): The description of the simulator model
    """

    def __init__(
        self,
        external_id: str,
        simulator_external_id: str,
        data_set_id: int,
        name: str,
        type: str,
        description: str | None = None,
    ) -> None:
        self.external_id = external_id
        self.simulator_external_id = simulator_external_id
        self.data_set_id = data_set_id
        self.name = name
        self.type = type
        self.description = description

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            simulator_external_id=resource["simulatorExternalId"],
            data_set_id=resource["dataSetId"],
            name=resource["name"],
            type=resource["type"],
            description=resource.get("description"),
        )


class SimulatorModelWrite(SimulatorModelCore):
    def as_write(self) -> SimulatorModelWrite:
        return self


class SimulatorModel(SimulatorModelCore):
    """
    The simulator model resource represents an asset modeled in a simulator.

    This asset could range from a pump or well to a complete processing facility or refinery.
    The simulator model is the root of its associated revisions, routines, runs, and results.
    The dataset assigned to a model is inherited by its children. Deleting a model also deletes all its children, thereby
    maintaining the integrity and hierarchy of the simulation data.
    Simulator model revisions track changes and updates to a simulator model over time.
    Each revision ensures that modifications to models are traceable and allows users to understand the evolution of a given model.

    This is the read/response format of a simulator model.

    Args:
        id (int): A unique id of a simulator model
        external_id (str): External id of the simulator model
        simulator_external_id (str): External id of the associated simulator
        data_set_id (int): The id of the dataset associated with the simulator model
        name (str): The name of the simulator model
        type (str): The type key of the simulator model
        created_time (int): The time when the simulator model was created
        last_updated_time (int): The time when the simulator model was last updated
        description (str | None): The description of the simulator model
    """

    def __init__(
        self,
        id: int,
        external_id: str,
        simulator_external_id: str,
        data_set_id: int,
        name: str,
        type: str,
        created_time: int,
        last_updated_time: int,
        description: str | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            simulator_external_id=simulator_external_id,
            data_set_id=data_set_id,
            name=name,
            type=type,
            description=description,
        )

        self.id = id
        self.created_time = created_time
        self.last_updated_time = last_updated_time

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            id=resource["id"],
            external_id=resource["externalId"],
            simulator_external_id=resource["simulatorExternalId"],
            data_set_id=resource["dataSetId"],
            name=resource["name"],
            type=resource["type"],
            description=resource.get("description"),
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
        )

    def as_write(self) -> SimulatorModelWrite:
        """Returns this SimulatorModel in its writing version."""
        return SimulatorModelWrite(
            external_id=self.external_id,
            simulator_external_id=self.simulator_external_id,
            data_set_id=self.data_set_id,
            name=self.name,
            type=self.type,
            description=self.description,
        )


class SimulatorModelWriteList(CogniteResourceList[SimulatorModelWrite], ExternalIDTransformerMixin):
    _RESOURCE = SimulatorModelWrite


class SimulatorModelList(WriteableCogniteResourceList[SimulatorModelWrite, SimulatorModel], IdTransformerMixin):
    _RESOURCE = SimulatorModel

    def as_write(self) -> SimulatorModelWriteList:
        return SimulatorModelWriteList([a.as_write() for a in self.data], cognite_client=self._get_cognite_client())


class SimulatorModelRevisionWriteList(CogniteResourceList[SimulatorModelRevisionWrite], ExternalIDTransformerMixin):
    _RESOURCE = SimulatorModelRevisionWrite


class SimulatorModelRevisionList(
    WriteableCogniteResourceList[SimulatorModelRevisionWrite, SimulatorModelRevision], IdTransformerMixin
):
    _RESOURCE = SimulatorModelRevision

    def as_write(self) -> SimulatorModelRevisionWriteList:
        return SimulatorModelRevisionWriteList(
            [a.as_write() for a in self.data], cognite_client=self._get_cognite_client()
        )


class SimulatorModelUpdate(CogniteUpdate):
    class _PrimitiveModelUpdate(CognitePrimitiveUpdate):
        def set(self, value: Any) -> None:
            self._set(value)

    @property
    def name(self) -> _PrimitiveModelUpdate:
        return SimulatorModelUpdate._PrimitiveModelUpdate(self, "name")

    @property
    def description(self) -> _PrimitiveModelUpdate:
        return SimulatorModelUpdate._PrimitiveModelUpdate(self, "description")

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            PropertySpec("name"),
            PropertySpec("description"),
        ]


@dataclass
class SimulatorModelDependencyFileReference(CogniteObject, ABC): ...


@dataclass
class SimulatorModelDependencyFileId(SimulatorModelDependencyFileReference):
    id: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(id=resource["id"])


@dataclass
class SimulatorModelRevisionDependency(CogniteObject):
    """
    Represents an external dependency for a simulator model revision.
    Args:
        file (int): The file ID associated with the external dependency.
        arguments (dict[str, str]): A dictionary that contains the key-value pairs (fields) for the external dependency.
    """

    file: SimulatorModelDependencyFileReference
    arguments: dict[str, str]

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            file=SimulatorModelDependencyFileId.load(resource["file"])
            if "id" in resource["file"]
            else resource["file"],
            arguments=resource["arguments"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        output["file"] = (
            self.file.dump(camel_case=camel_case)
            if isinstance(self.file, SimulatorModelDependencyFileReference)
            else self.file
        )
        return output

    @classmethod
    def _load_list(
        cls, resource: list[dict[str, Any]], cognite_client: AsyncCogniteClient | None = None
    ) -> list[SimulatorModelRevisionDependency]:
        return [cls._load(item, cognite_client) for item in resource]


@dataclass
class SimulatorFlowsheetObjectEdge(CogniteObject):
    id: str
    name: str | None
    source_id: str
    target_id: str
    connection_type: Literal["Material", "Energy", "Information"]

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            id=resource["id"],
            name=resource.get("name"),
            source_id=resource["sourceId"],
            target_id=resource["targetId"],
            connection_type=resource["connectionType"],
        )

    @classmethod
    def _load_list(
        cls, resource: list[dict[str, Any]], cognite_client: AsyncCogniteClient | None = None
    ) -> list[SimulatorFlowsheetObjectEdge]:
        return [cls._load(item, cognite_client) for item in resource]


@dataclass
class SimulatorFlowsheetThermodynamic(CogniteObject):
    property_packages: list[str]
    components: list[str]

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            property_packages=resource["propertyPackages"],
            components=resource["components"],
        )


@dataclass
class SimulationValueUnitReference(CogniteObject):
    name: str
    quantity: str | None = None
    external_id: str | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            quantity=resource.get("quantity"),
            external_id=resource.get("externalId"),
        )


@dataclass
class SimulatorFlowsheetProperty(CogniteObject):
    name: str
    reference_object: dict[str, str]
    value_type: Literal["STRING", "DOUBLE", "STRING_ARRAY", "DOUBLE_ARRAY"]
    value: str | float | list[str] | list[float]
    unit: SimulationValueUnitReference | None
    read_only: bool | None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            name=resource["name"],
            reference_object=resource["referenceObject"],
            value_type=resource["valueType"],
            value=resource["value"],
            unit=SimulationValueUnitReference._load(resource["unit"]) if "unit" in resource else None,
            read_only=resource.get("readOnly"),
        )

    @classmethod
    def _load_list(
        cls, resource: list[dict[str, Any]], cognite_client: AsyncCogniteClient | None = None
    ) -> list[SimulatorFlowsheetProperty]:
        return [cls._load(item, cognite_client) for item in resource]

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.unit is not None:
            output["unit"] = self.unit.dump(camel_case=camel_case)

        return output


@dataclass
class SimulatorFlowsheetPosition(CogniteObject):
    x: float
    y: float

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            x=resource["x"],
            y=resource["y"],
        )


@dataclass
class SimulatorFlowsheetGraphicalObject(CogniteObject):
    position: SimulatorFlowsheetPosition | None
    height: float | None
    width: float | None
    scale_x: float | int | None
    scale_y: float | int | None
    angle: float | None
    active: bool | None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            position=SimulatorFlowsheetPosition._load(resource["position"]) if "position" in resource else None,
            height=resource.get("height"),
            width=resource.get("width"),
            scale_x=resource.get("scaleX"),
            scale_y=resource.get("scaleY"),
            angle=resource.get("angle"),
            active=resource.get("active"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.position is not None:
            output["position"] = self.position.dump(camel_case=camel_case)

        return output


@dataclass
class SimulatorFlowsheetObjectNode(CogniteObject):
    id: str
    name: str | None
    type: str
    graphical_object: SimulatorFlowsheetGraphicalObject | None
    properties: list[SimulatorFlowsheetProperty]

    @classmethod
    def _load_list(
        cls, resource: list[dict[str, Any]], cognite_client: AsyncCogniteClient | None = None
    ) -> list[SimulatorFlowsheetObjectNode]:
        return [cls._load(item, cognite_client) for item in resource]

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            id=resource["id"],
            name=resource.get("name"),
            type=resource["type"],
            graphical_object=SimulatorFlowsheetGraphicalObject._load(resource["graphicalObject"])
            if "graphicalObject" in resource
            else None,
            properties=SimulatorFlowsheetProperty._load_list(resource["properties"], cognite_client),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.graphical_object is not None:
            output["graphicalObject"] = self.graphical_object.dump(camel_case=camel_case)
        output["properties"] = [item.dump(camel_case=camel_case) for item in self.properties]

        return output


@dataclass
class SimulatorFlowsheet(CogniteObject):
    simulator_object_nodes: list[SimulatorFlowsheetObjectNode]
    simulator_object_edges: list[SimulatorFlowsheetObjectEdge]
    thermodynamics: SimulatorFlowsheetThermodynamic

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            simulator_object_nodes=SimulatorFlowsheetObjectNode._load_list(
                resource["simulatorObjectNodes"], cognite_client
            ),
            simulator_object_edges=SimulatorFlowsheetObjectEdge._load_list(
                resource["simulatorObjectEdges"], cognite_client
            ),
            thermodynamics=SimulatorFlowsheetThermodynamic._load(resource["thermodynamics"], cognite_client),
        )

    @classmethod
    def _load_list(
        cls, resource: list[dict[str, Any]], cognite_client: AsyncCogniteClient | None = None
    ) -> list[SimulatorFlowsheet]:
        return [cls._load(item, cognite_client) for item in resource]

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        output["simulatorObjectNodes"] = [item.dump(camel_case=camel_case) for item in self.simulator_object_nodes]
        output["simulatorObjectEdges"] = [item.dump(camel_case=camel_case) for item in self.simulator_object_edges]
        if self.thermodynamics is not None:
            output["thermodynamics"] = self.thermodynamics.dump(camel_case=camel_case)

        return output


@dataclass
class SimulatorModelRevisionData(CogniteResource):
    """
    Extracted metadata from a simulator model file associated with a model revision.

    When a model revision is created, connectors can optionally parse the simulator file
    to extract structured information about the model's internal structure and configuration.
    This data resource stores the parsed information, which may include flowsheet details,
    process equipment, operating parameters, connections between blocks, and visualization data.

    Note: The availability and extent of this data depends entirely on the connector
    implementation and simulator type. Some connectors may:
    - Not implement this feature at all (no data extraction)
    - Partially implement it (e.g., only populate 'info' or only 'flowsheets')
    - Fully implement it with comprehensive model details

    Args:
        model_revision_external_id (str): External id of the associated model revision
        created_time (int): The time when the simulator model revision data was created
        last_updated_time (int): The time when the simulator model revision data was last updated
        data_set_id (int): The id of the dataset associated with the simulator model revision data
        flowsheets (list[SimulatorFlowsheet] | None): Extracted flowsheet information,
            if supported by the connector. May include blocks, equipment, properties, and connections
        info (dict[str, str] | None): Additional metadata extracted from the simulator file,
            if supported by the connector
    """

    def __init__(
        self,
        model_revision_external_id: str,
        created_time: int,
        last_updated_time: int,
        data_set_id: int,
        flowsheets: list[SimulatorFlowsheet] | None,
        info: dict[str, str] | None,
    ) -> None:
        self.model_revision_external_id = model_revision_external_id
        self.flowsheets = flowsheets
        self.info = info
        self.data_set_id = data_set_id
        self.created_time = created_time
        self.last_updated_time = last_updated_time

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            model_revision_external_id=resource["modelRevisionExternalId"],
            flowsheets=SimulatorFlowsheet._load_list(resource["flowsheets"], cognite_client)
            if resource.get("flowsheets")
            else None,
            info=resource.get("info"),
            data_set_id=resource["dataSetId"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.flowsheets is not None:
            output["flowsheets"] = [
                item.dump(camel_case=camel_case) if isinstance(item, SimulatorFlowsheet) else item
                for item in self.flowsheets
            ]

        return output


class SimulatorModelRevisionDataList(CogniteResourceList[SimulatorModelRevisionData], ExternalIDTransformerMixin):
    _RESOURCE = SimulatorModelRevisionData



from __future__ import annotations

from collections.abc import AsyncIterator, Sequence
from typing import TYPE_CHECKING, Any, Literal, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.hosted_extractors.destinations import (
    Destination,
    DestinationList,
    DestinationUpdate,
    DestinationWrite,
)
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient, ClientConfig


class DestinationsAPI(APIClient):
    _RESOURCE_PATH = "/hostedextractors/destinations"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: AsyncCogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self._warning = FeaturePreviewWarning(
            api_maturity="beta", sdk_maturity="alpha", feature_name="Hosted Extractors"
        )
        self._CREATE_LIMIT = 10
        self._LIST_LIMIT = 100
        self._RETRIEVE_LIMIT = 100
        self._DELETE_LIMIT = 100
        self._UPDATE_LIMIT = 10

    @overload
    def __call__(self, chunk_size: None = None, limit: int | None = None) -> AsyncIterator[Destination]: ...

    @overload
    def __call__(self, chunk_size: int, limit: int | None = None) -> AsyncIterator[Destination]: ...

    async def __call__(
        self, chunk_size: int | None = None, limit: int | None = None
    ) -> AsyncIterator[Destination] | AsyncIterator[DestinationList]:
        """Iterate over destinations

        Fetches Destination as they are iterated over, so you keep a limited number of destinations in memory.

        Args:
            chunk_size (int | None): Number of Destinations to return in each chunk. Defaults to yielding one Destination a time.
            limit (int | None): Maximum number of Destination to return. Defaults to returning all items.

        Yields:
            Destination | DestinationList: yields Destination one by one if chunk_size is not specified, else DestinationList objects.
        """  # noqa: DOC404
        self._warning.warn()

        async for item in self._list_generator(
            list_cls=DestinationList,
            resource_cls=Destination,
            method="GET",
            chunk_size=chunk_size,
            limit=limit,
            headers={"cdf-version": "beta"},
        ):
            yield item

    @overload
    async def retrieve(self, external_ids: str, ignore_unknown_ids: bool = False) -> Destination: ...

    @overload
    async def retrieve(
        self, external_ids: SequenceNotStr[str], ignore_unknown_ids: bool = False
    ) -> DestinationList: ...

    async def retrieve(
        self, external_ids: str | SequenceNotStr[str], ignore_unknown_ids: bool = False
    ) -> Destination | DestinationList:
        """`Retrieve one or more destinations. <https://api-docs.cognite.com/20230101-beta/tag/Destinations/operation/retrieve_destinations>`_

        Args:
            external_ids (str | SequenceNotStr[str]): The external ID provided by the client. Must be unique for the resource type.
            ignore_unknown_ids (bool): Ignore external IDs that are not found


        Returns:
            Destination | DestinationList: Requested destinations

        Examples:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.hosted_extractors.destinations.retrieve('myDestination')

            Get multiple destinations by id:

                >>> res = client.hosted_extractors.destinations.retrieve(["myDestination", "myDestination2"], ignore_unknown_ids=True)

        """
        self._warning.warn()
        return await self._retrieve_multiple(
            list_cls=DestinationList,
            resource_cls=Destination,
            identifiers=IdentifierSequence.load(external_ids=external_ids),
            ignore_unknown_ids=ignore_unknown_ids,
            headers={"cdf-version": "beta"},
        )

    async def delete(
        self, external_ids: str | SequenceNotStr[str], ignore_unknown_ids: bool = False, force: bool = False
    ) -> None:
        """`Delete one or more destsinations <https://api-docs.cognite.com/20230101-beta/tag/Destinations/operation/delete_destinations>`_

        Args:
            external_ids (str | SequenceNotStr[str]): The external ID provided by the client. Must be unique for the resource type.
            ignore_unknown_ids (bool): Ignore external IDs that are not found
            force (bool): Delete any jobs associated with each item.

        Examples:

            Delete destinations by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> client.hosted_extractors.destinations.delete(["myDest", "MyDest2"])
        """
        self._warning.warn()
        extra_body_fields: dict[str, Any] = {}
        if ignore_unknown_ids:
            extra_body_fields["ignoreUnknownIds"] = True
        if force:
            extra_body_fields["force"] = True

        await self._delete_multiple(
            identifiers=IdentifierSequence.load(external_ids=external_ids),
            wrap_ids=True,
            returns_items=False,
            headers={"cdf-version": "beta"},
            extra_body_fields=extra_body_fields or None,
        )

    @overload
    async def create(self, items: DestinationWrite) -> Destination: ...

    @overload
    async def create(self, items: Sequence[DestinationWrite]) -> DestinationList: ...

    async def create(self, items: DestinationWrite | Sequence[DestinationWrite]) -> Destination | DestinationList:
        """`Create one or more destinations. <https://api-docs.cognite.com/20230101-beta/tag/Destinations/operation/create_destinations>`_

        Args:
            items (DestinationWrite | Sequence[DestinationWrite]): Destination(s) to create.

        Returns:
            Destination | DestinationList: Created destination(s)

        Examples:

            Create new destination:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.hosted_extractors import DestinationWrite, SessionWrite
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> destination = DestinationWrite(external_id='my_dest', credentials=SessionWrite("my_nonce"), target_data_set_id=123)
                >>> res = client.hosted_extractors.destinations.create(destination)
        """
        self._warning.warn()
        return await self._create_multiple(
            list_cls=DestinationList,
            resource_cls=Destination,
            items=items,
            input_resource_cls=DestinationWrite,
            headers={"cdf-version": "beta"},
        )

    @overload
    async def update(
        self,
        items: DestinationWrite | DestinationUpdate,
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> Destination: ...

    @overload
    async def update(
        self,
        items: Sequence[DestinationWrite | DestinationUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> DestinationList: ...

    async def update(
        self,
        items: DestinationWrite | DestinationUpdate | Sequence[DestinationWrite | DestinationUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> Destination | DestinationList:
        """`Update one or more destinations. <https://api-docs.cognite.com/20230101-beta/tag/Destinations/operation/update_destinations>`_

        Args:
            items (DestinationWrite | DestinationUpdate | Sequence[DestinationWrite | DestinationUpdate]): Destination(s) to update.
            mode (Literal['replace_ignore_null', 'patch', 'replace']): How to update data when a non-update object is given (DestinationWrite). If you use 'replace_ignore_null', only the fields you have set will be used to replace existing (default). Using 'replace' will additionally clear all the fields that are not specified by you. Last option, 'patch', will update only the fields you have set and for container-like fields such as metadata or labels, add the values to the existing. For more details, see :ref:`appendix-update`.

        Returns:
            Destination | DestinationList: Updated destination(s)

        Examples:

            Update destination:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.hosted_extractors import DestinationUpdate
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> destination = DestinationUpdate('my_dest').target_data_set_id.set(123)
                >>> res = client.hosted_extractors.destinations.update(destination)
        """
        self._warning.warn()
        return await self._update_multiple(
            items=items,
            list_cls=DestinationList,
            resource_cls=Destination,
            update_cls=DestinationUpdate,
            mode=mode,
            headers={"cdf-version": "beta"},
        )

    async def list(
        self,
        limit: int | None = DEFAULT_LIMIT_READ,
    ) -> DestinationList:
        """`List destinations <https://api-docs.cognite.com/20230101-beta/tag/Destinations/operation/list_destinations>`_

        Args:
            limit (int | None): Maximum number of destinations to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            DestinationList: List of requested destinations

        Examples:

            List destinations:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> destination_list = client.hosted_extractors.destinations.list(limit=5)

            Iterate over destinations, one-by-one:

                >>> for destination in client.hosted_extractors.destinations():
                ...     destination  # do something with the destination

            Iterate over chunks of destinations to reduce memory load:

                >>> for destination_list in client.hosted_extractors.destinations(chunk_size=25):
                ...     destination_list # do something with the destinationss
        """
        self._warning.warn()
        return await self._list(
            list_cls=DestinationList,
            resource_cls=Destination,
            method="GET",
            limit=limit,
            headers={"cdf-version": "beta"},
        )



from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Literal, cast, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes import (
    ExtractionPipelineRun,
    ExtractionPipelineRunFilter,
    ExtractionPipelineRunList,
    TimestampRange,
)
from cognite.client.data_classes.extractionpipelines import (
    ExtractionPipelineRunCore,
    ExtractionPipelineRunWrite,
    StringFilter,
)
from cognite.client.utils import timestamp_to_ms
from cognite.client.utils._validation import assert_type
from cognite.client.utils.useful_types import SequenceNotStr

RunStatus = Literal["success", "failure", "seen"]


class ExtractionPipelineRunsAPI(APIClient):
    _RESOURCE_PATH = "/extpipes/runs"

    async def list(
        self,
        external_id: str,
        statuses: RunStatus | Sequence[RunStatus] | SequenceNotStr[str] | None = None,
        message_substring: str | None = None,
        created_time: dict[str, Any] | TimestampRange | str | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
    ) -> ExtractionPipelineRunList:
        """`List runs for an extraction pipeline with given external_id <https://developer.cognite.com/api#tag/Extraction-Pipelines-Runs/operation/filterRuns>`_

        Args:
            external_id (str): Extraction pipeline external Id.
            statuses (RunStatus | Sequence[RunStatus] | SequenceNotStr[str] | None): One or more among "success" / "failure" / "seen".
            message_substring (str | None): Failure message part.
            created_time (dict[str, Any] | TimestampRange | str | None): Range between two timestamps. Possible keys are `min` and `max`, with values given as timestamps in ms.
                If a string is passed, it is assumed to be the minimum value.
            limit (int | None): Maximum number of ExtractionPipelines to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            ExtractionPipelineRunList: List of requested extraction pipeline runs

        Tip:
            The ``created_time`` parameter can also be passed as a string, to support the most typical usage pattern
            of fetching the most recent runs, meaning it is implicitly assumed to be the minimum created time. The
            format is "N[timeunit]-ago", where timeunit is w,d,h,m (week, day, hour, minute), e.g. "12d-ago".

        Examples:

            List extraction pipeline runs:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> runsList = client.extraction_pipelines.runs.list(external_id="test ext id", limit=5)

            Filter extraction pipeline runs on a given status:

                >>> runs_list = client.extraction_pipelines.runs.list(external_id="test ext id", statuses=["seen"], limit=5)

            Get all failed pipeline runs in the last 24 hours for pipeline 'extId':

                >>> from cognite.client.data_classes import ExtractionPipelineRun
                >>> res = client.extraction_pipelines.runs.list(external_id="extId", statuses="failure", created_time="24h-ago")
        """
        if isinstance(created_time, str):
            created_time = TimestampRange(min=timestamp_to_ms(created_time))

        if statuses is not None or message_substring is not None or created_time is not None:
            filter = ExtractionPipelineRunFilter(
                external_id=external_id,
                statuses=cast(SequenceNotStr[str] | None, [statuses] if isinstance(statuses, str) else statuses),
                message=StringFilter(substring=message_substring),
                created_time=created_time,
            ).dump(camel_case=True)
            method: Literal["POST", "GET"] = "POST"
        else:
            filter = {"externalId": external_id}
            method = "GET"

        res = await self._list(
            list_cls=ExtractionPipelineRunList,
            resource_cls=ExtractionPipelineRun,
            method=method,
            limit=limit,
            filter=filter,
        )
        for run in res:
            run.extpipe_external_id = external_id
        return res

    @overload
    async def create(self, run: ExtractionPipelineRun | ExtractionPipelineRunWrite) -> ExtractionPipelineRun: ...

    @overload
    async def create(
        self, run: Sequence[ExtractionPipelineRun] | Sequence[ExtractionPipelineRunWrite]
    ) -> ExtractionPipelineRunList: ...

    async def create(
        self,
        run: ExtractionPipelineRun
        | ExtractionPipelineRunWrite
        | Sequence[ExtractionPipelineRun]
        | Sequence[ExtractionPipelineRunWrite],
    ) -> ExtractionPipelineRun | ExtractionPipelineRunList:
        """`Create one or more extraction pipeline runs. <https://developer.cognite.com/api#tag/Extraction-Pipelines-Runs/operation/createRuns>`_

        You can create an arbitrary number of extraction pipeline runs, and the SDK will split the request into multiple requests.

        Args:
            run (ExtractionPipelineRun | ExtractionPipelineRunWrite | Sequence[ExtractionPipelineRun] | Sequence[ExtractionPipelineRunWrite]): ExtractionPipelineRun| ExtractionPipelineRunWrite | Sequence[ExtractionPipelineRun]  | Sequence[ExtractionPipelineRunWrite]): Extraction pipeline or list of extraction pipeline runs to create.

        Returns:
            ExtractionPipelineRun | ExtractionPipelineRunList: Created extraction pipeline run(s)

        Examples:

            Report a new extraction pipeline run:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import ExtractionPipelineRunWrite
                >>> client = CogniteClient()
                >>> res = client.extraction_pipelines.runs.create(
                ...     ExtractionPipelineRunWrite(status="success", extpipe_external_id="extId"))
        """
        assert_type(run, "run", [ExtractionPipelineRunCore, Sequence])
        return await self._create_multiple(
            list_cls=ExtractionPipelineRunList,
            resource_cls=ExtractionPipelineRun,
            items=run,
            input_resource_cls=ExtractionPipelineRunWrite,
        )



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Any, TypeVar, cast, overload

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteResourceList,
    NameTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.utils._async_helpers import run_sync
from cognite.client.utils._importing import local_import
from cognite.client.utils._text import copy_doc_from_async

if TYPE_CHECKING:
    import pandas

    from cognite.client import AsyncCogniteClient


class RowCore(WriteableCogniteResource["RowWrite"], ABC):
    """No description.

    Args:
        key (str): Unique row key
        columns (dict[str, Any]): Row data stored as a JSON object.
    """

    def __init__(
        self,
        key: str,
        columns: dict[str, Any],
    ) -> None:
        self.key = key
        self.columns = columns

    def get(self, attr: str, default: Any = None) -> Any:
        return (self.columns or {}).get(attr, default)

    def __getitem__(self, attr: str) -> Any:
        return (self.columns or {})[attr]

    def __setitem__(self, attr: str, value: Any) -> None:
        if self.columns is not None:
            self.columns[attr] = value
        else:
            raise RuntimeError("columns not set on Row instance")

    def __delitem__(self, attr: str) -> None:
        if self.columns is not None:
            del self.columns[attr]
        else:
            raise RuntimeError("columns not set on Row instance")

    def __contains__(self, attr: str) -> bool:
        return self.columns is not None and attr in self.columns

    def to_pandas(self) -> pandas.DataFrame:  # type: ignore[override]
        """Convert the instance into a pandas DataFrame.

        Returns:
            pandas.DataFrame: The pandas DataFrame representing this instance.
        """
        pd = local_import("pandas")
        return pd.DataFrame([self.columns], [self.key])


T_Row = TypeVar("T_Row", bound=RowCore)


class Row(RowCore):
    """This represents a row in a NO-SQL table.
    This is the reading version of the Row class, which is used when retrieving a row.

    Args:
        key (str): Unique row key
        columns (dict[str, Any]): Row data stored as a JSON object.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        cognite_client (AsyncCogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        key: str,
        columns: dict[str, Any],
        last_updated_time: int,
        cognite_client: AsyncCogniteClient | None = None,
    ) -> None:
        super().__init__(key, columns)
        self.last_updated_time = last_updated_time
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            key=resource["key"],
            columns=resource["columns"],
            last_updated_time=resource["lastUpdatedTime"],
            cognite_client=cognite_client,
        )

    def as_write(self) -> RowWrite:
        """Returns this Row as a RowWrite"""
        if self.key is None or self.columns is None:
            raise ValueError("key and columns are required to create a Row")
        return RowWrite(key=self.key, columns=self.columns)


class RowWrite(RowCore):
    """This represents a row in a NO-SQL table.
    This is the writing version of the Row class, which is used when creating a row.

    Args:
        key (str): Unique row key
        columns (dict[str, Any]): Row data stored as a JSON object.
    """

    def __init__(self, key: str, columns: dict[str, Any]) -> None:
        super().__init__(key, columns)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> RowWrite:
        return cls(resource["key"], resource["columns"])

    def as_write(self) -> RowWrite:
        """Returns this RowWrite instance."""
        return self


class RowListCore(WriteableCogniteResourceList[RowWrite, T_Row], ABC):
    def to_pandas(self) -> pandas.DataFrame:  # type: ignore[override]
        """Convert the instance into a pandas DataFrame.

        Returns:
            pandas.DataFrame: The pandas DataFrame representing this instance.
        """
        pd = local_import("pandas")
        if not self:
            return pd.DataFrame(columns=[], index=[])
        index, data = zip(*((row.key, row.columns) for row in self))
        return pd.DataFrame.from_records(data, index=index)


class RowWriteList(RowListCore[RowWrite]):
    _RESOURCE = RowWrite

    def as_write(self) -> RowWriteList:
        return self


class RowList(RowListCore[Row]):
    _RESOURCE = Row

    def as_write(self) -> RowWriteList:
        """Returns this RowList as a RowWriteList"""
        return RowWriteList([row.as_write() for row in self.data])


class TableCore(WriteableCogniteResource["TableWrite"]):
    """A NoSQL database table to store customer data

    Args:
        name (str): Unique name of the table
    """

    def __init__(self, name: str) -> None:
        self.name = name


class Table(TableCore):
    """A NoSQL database table to store customer data.
    This is the reading version of the Table class, which is used when retrieving a table.

    Args:
        name (str): Unique name of the table
        created_time (int | None): Time the table was created.
        cognite_client (AsyncCogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        name: str,
        created_time: int | None,
        cognite_client: AsyncCogniteClient | None = None,
    ) -> None:
        super().__init__(name)
        self.created_time = created_time
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)

        self._db_name: str | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(name=resource["name"], created_time=resource.get("createdTime"), cognite_client=cognite_client)

    def as_write(self) -> TableWrite:
        """Returns this Table as a TableWrite"""
        if self.name is None:
            raise ValueError("name is required to create a Table")
        return TableWrite(name=self.name)

    @overload
    async def rows_async(self, key: str, limit: int | None = None) -> Row | None: ...

    @overload
    async def rows_async(self, key: None = None, limit: int | None = None) -> RowList: ...

    async def rows_async(self, key: str | None = None, limit: int | None = None) -> Row | RowList | None:
        """Get the rows in this table.

        Args:
            key (str | None): Specify a key to return only that row.
            limit (int | None): The number of rows to return.

        Returns:
            Row | RowList | None: List of tables in this database.
        """
        if self._db_name is None:
            raise ValueError("Table is not linked to a database, did you instantiate it yourself?")
        elif self.name is None:
            raise ValueError("Table 'name' is missing")

        if key is not None:
            return await self._cognite_client.raw.rows.retrieve(db_name=self._db_name, table_name=self.name, key=key)
        return await self._cognite_client.raw.rows.list(db_name=self._db_name, table_name=self.name, limit=limit)

    @overload
    def rows(self, key: str, limit: int | None = None) -> Row | None: ...

    @overload
    def rows(self, key: None = None, limit: int | None = None) -> RowList: ...

    @copy_doc_from_async(rows_async)
    def rows(self, key: str | None = None, limit: int | None = None) -> Row | RowList | None:
        return run_sync(self.rows_async(key=key, limit=limit))


class TableWrite(TableCore):
    """A NoSQL database table to store customer data
    This is the writing version of the Table class, which is used when creating a table.

    Args:
        name (str): Unique name of the table
    """

    def __init__(self, name: str) -> None:
        super().__init__(name)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> TableWrite:
        return cls(resource["name"])

    def as_write(self) -> TableWrite:
        """Returns this TableWrite instance."""
        return self


class TableWriteList(CogniteResourceList[TableWrite], NameTransformerMixin):
    _RESOURCE = TableWrite


class TableList(WriteableCogniteResourceList[TableWrite, Table], NameTransformerMixin):
    _RESOURCE = Table

    def as_write(self) -> TableWriteList:
        """Returns this TableList as a TableWriteList"""
        return TableWriteList([table.as_write() for table in self.data])


class DatabaseCore(WriteableCogniteResource["DatabaseWrite"], ABC):
    """A NoSQL database to store customer data.

    Args:
        name (str): Unique name of a database.
    """

    def __init__(self, name: str) -> None:
        self.name = name


class Database(DatabaseCore):
    """A NoSQL database to store customer data.

    Args:
        name (str): Unique name of a database.
        created_time (int | None): Time the database was created.
        cognite_client (AsyncCogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        name: str,
        created_time: int | None,
        cognite_client: AsyncCogniteClient | None = None,
    ) -> None:
        super().__init__(name)
        self.created_time = created_time
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(name=resource["name"], created_time=resource.get("createdTime"), cognite_client=cognite_client)

    def as_write(self) -> DatabaseWrite:
        """Returns this Database as a DatabaseWrite"""
        if self.name is None:
            raise ValueError("name is required to create a Database")
        return DatabaseWrite(name=self.name)

    async def tables_async(self, limit: int | None = None) -> TableList:
        """Get the tables in this database.

        Args:
            limit (int | None): The number of tables to return.

        Returns:
            TableList: List of tables in this database.
        """
        if self.name is None:
            raise ValueError("Unable to list tables, 'name' is not set on instance")
        return await self._cognite_client.raw.tables.list(db_name=self.name, limit=limit)

    @copy_doc_from_async(tables_async)
    def tables(self, limit: int | None = None) -> TableList:
        return run_sync(self.tables_async(limit=limit))


class DatabaseWrite(DatabaseCore):
    """A NoSQL database to store customer data.

    Args:
        name (str): Unique name of a database.
    """

    def __init__(self, name: str) -> None:
        super().__init__(name)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> DatabaseWrite:
        return cls(resource["name"])

    def as_write(self) -> DatabaseWrite:
        """Returns this DatabaseWrite instance."""
        return self


class DatabaseWriteList(CogniteResourceList[DatabaseWrite], NameTransformerMixin):
    _RESOURCE = DatabaseWrite


class DatabaseList(WriteableCogniteResourceList[DatabaseWrite, Database], NameTransformerMixin):
    _RESOURCE = Database

    def as_write(self) -> DatabaseWriteList:
        """Returns this DatabaseList as a DatabaseWriteList"""
        return DatabaseWriteList([db.as_write() for db in self.data])
